{ File   : DEFParser.pas
  Author : Deniz Oezmen
  Created: 2005-03-06
  Changed: 2006-08-13

  Pretty stupid DEF scanner/parser for ROML scene files.
}
unit DEFParser;

interface

uses
  DEFTypes;

const
  // return codes
  DEFParser_Success     = Longint($00000000);
  DEFParser_FileNotOpen = Longint($80010001);
  DEFParser_ParserError = Longint($80010002);

function ParseDEF(DEFName: string; var DEF: TDEF): Longint;

implementation

uses
  SysUtils;

const
  // token classes
  NoMatch    = 0;
  Identifier = 1;
  Numeral    = 2;
  Colon      = 3;
  LiteralStr = 4;
  BrckOp     = 5;
  BrckCl     = 6;
  CrlBrckOp  = 7;
  CrlBrckCl  = 8;
  Comma      = 9;
  Eoi        = 10;

var
  DEFFile   : file of Char;
  Token     : Longword;
  Value     : string;

function OpenDEF(DEFName: string): Longint;
begin
  AssignFile(DEFFile, DEFName);
  FileMode := fmOpenRead;
  {$I-}
  Reset(DEFFile);
  {$I+}
  if (IOResult = 0) then
    Result := DEFParser_Success
  else
    Result := DEFParser_FileNotOpen
end;

(* naive scanner routine *)
function GetNextSymbol(var Token: Longword; var Value: string): Boolean;
const
  // character sets
  Alpha      = ['a'..'z', 'A'..'Z', '_'];
  Num        = ['0'..'9'];
  AlphaNum   = Alpha + Num;
  Blank      = [' ', #009, #010, #013];
  Delim      = [':', ',', '(', ')', '{', '}', '"'];
var
  c: Char;
begin
  Token := NoMatch;
  Value := '';
  Result := False;

  while True do
  begin
    if not Eof(DEFFile) then
      Read(DEFFile, c)
    else
    begin
      Result := (Value <> '') and (Token <> LiteralStr);
      Token := Eoi;
      Break
    end;

    case Token of
      NoMatch:
        if c in Alpha then
        begin
          Token := Identifier;
          Value := Value + c
        end
        else if (c in Num) or (c = '-') then
        begin
          Token := Numeral;
          Value := Value + c
        end
        else if c = ':' then
        begin
          Token := Colon;
          Value := Value + c;
          Result := True;
          Break
        end
        else if c = '"' then
          Token := LiteralStr
        else if c = '{' then
        begin
          Token := CrlBrckOp;
          Value := Value + c;
          Result := True;
          Break
        end
        else if c = '}' then
        begin
          Token := CrlBrckCl;
          Value := Value + c;
          Result := True;
          Break
        end
        else if c = '(' then
        begin
          Token := BrckOp;
          Value := Value + c;
          Result := True;
          Break
        end
        else if c = ')' then
        begin
          Token := BrckCl;
          Value := Value + c;
          Result := True;
          Break
        end
        else if c = ',' then
        begin
          Token := Comma;
          Value := Value + c;
          Result := True;
          Break
        end
        else if not (c in Blank) then
          Break; 
      Identifier:
        if c in AlphaNum then
          Value := Value + c
        else
        begin
          Seek(DEFFile, FilePos(DEFFile) - 1);
          Result := True;
          Break
        end;
      Numeral:
        if c in Num then
          Value := Value + c
        else if c in Blank then
        begin
          Result := True;
          Break
        end
        else if c in Delim then
        begin
          Seek(DEFFile, FilePos(DEFFile) - 1);
          Result := True;
          Break
        end
        else
          Break;
      LiteralStr:
        if c = '"' then
        begin
          Result := True;
          Break;
        end
        else
          Value := Value + c
    end
  end;

  if not Result and IsConsole then
    WriteLn('Scanner error.')
end;

function GetScene(var Scene: string): Boolean;
begin
  Result := False;

  if Scene <> '' then
  begin
    if IsConsole then
      WriteLn('''scene'' redefined.');
    Exit
  end;
  if not GetNextSymbol(Token, Value) then
  begin
    if IsConsole then
      WriteLn(''':'' expected.');
    Exit
  end;
  if (not GetNextSymbol(Token, Value)) or ((Token <> Identifier)
    and (Token <> Numeral))then
  begin
    if IsConsole then
      WriteLn('Identifier or numeral expected.');
    Exit
  end;

  Scene := Value;
  Result := True
end;

function GetArtbase(var Artbase: string): Boolean;
begin
  Result := False;

  if Artbase <> '' then
  begin
    if IsConsole then
      WriteLn('''artbase'' redefined.');
    Exit
  end;
  if not GetNextSymbol(Token, Value) then
  begin
    if IsConsole then
      WriteLn(''':'' expected.');
    Exit
  end;
  if (not GetNextSymbol(Token, Value)) or (Token <> LiteralStr) then
  begin
    if IsConsole then
      WriteLn('Literal string expected.');
    Exit
  end;

  Artbase := Value;
  Result := True
end;

function GetFrontscale(var Frontscale: Longint): Boolean;
begin
  Result := False;

  if Frontscale <> -1 then
  begin
    if IsConsole then
      WriteLn('''frontscale'' redefined.');
    Exit
  end;
  if not GetNextSymbol(Token, Value) then
  begin
    if IsConsole then
      WriteLn(''':'' expected.');
    Exit
  end;
  if (not GetNextSymbol(Token, Value)) or (Token <> Numeral) then
  begin
    if IsConsole then
      WriteLn('Numeral expected.');
    Exit
  end;

  Frontscale := StrToInt(Value);
  Result := True
end;

function GetBackscale(var Backscale: Longint): Boolean;
begin
  Result := False;

  if Backscale <> -1 then
  begin
    if IsConsole then
      WriteLn('''backscale'' redefined.');
    Exit
  end;
  if not GetNextSymbol(Token, Value) then
  begin
    if IsConsole then
      WriteLn(''':'' expected.');
    Exit
  end;
  if (not GetNextSymbol(Token, Value)) or (Token <> Numeral) then
  begin
    if IsConsole then
      WriteLn('Numeral expected.');
    Exit
  end;

  Backscale := StrToInt(Value);
  Result := True
end;

function GetFronty(var Fronty: Longint): Boolean;
begin
  Result := False;

  if Fronty <> -1 then
  begin
    if IsConsole then
      WriteLn('''fronty'' redefined.');
    Exit
  end;
  if not GetNextSymbol(Token, Value) then
  begin
    if IsConsole then
      WriteLn(''':'' expected.');
    Exit
  end;
  if (not GetNextSymbol(Token, Value)) or (Token <> Numeral) then
  begin
    if IsConsole then
      WriteLn('Numeral expected.');
    Exit
  end;

  Fronty := StrToInt(Value);
  Result := True
end;

function GetBacky(var Backy: Longint): Boolean;
begin
  Result := False;

  if Backy <> -1 then
  begin
    if IsConsole then
      WriteLn('''backy'' redefined.');
    Exit
  end;
  if not GetNextSymbol(Token, Value) then
  begin
    if IsConsole then
      WriteLn(''':'' expected.');
    Exit
  end;
  if (not GetNextSymbol(Token, Value)) or (Token <> Numeral) then
  begin
    if IsConsole then
      WriteLn('Numeral expected.');
    Exit
  end;

  Backy := StrToInt(Value);
  Result := True
end;

function GetTuple(var Rails: TRails): Boolean;
var
  Tuple: TTuple;
begin
  Result := False;

  if (not GetNextSymbol(Token, Value)) or (Token <> Numeral) then
  begin
    if IsConsole then
      WriteLn('Numeral expected.');
    Exit
  end;
  Tuple.v1 := StrToInt(Value);

  if (not GetNextSymbol(Token, Value)) or (Token <> Comma) then
  begin
    if IsConsole then
      WriteLn(''','' expected.');
    Exit
  end;

  if (not GetNextSymbol(Token, Value)) or (Token <> Numeral) then
  begin
    if IsConsole then
      WriteLn('Numeral expected.');
    Exit
  end;
  Tuple.v2 := StrToInt(Value);

  if (not GetNextSymbol(Token, Value)) or (Token <> BrckCl) then
  begin
    if IsConsole then
      WriteLn(''')'' expected.');
    Exit
  end;

  SetLength(Rails, Length(Rails) + 1);
  Rails[High(Rails)] := Tuple;
  Result := True
end;

function GetRails(var Rails: TRails): Boolean;
begin
  Result := False;

  if Length(Rails) > 0 then
  begin
    if IsConsole then
      WriteLn('''rails'' redefined.');
    Exit
  end;

  if (not GetNextSymbol(Token, Value)) or (Token <> CrlBrckOp) then
  begin
    if IsConsole then
      WriteLn('''{'' expected.');
    Exit
  end;

  while True do
  begin
    if not GetNextSymbol(Token, Value) then
    begin
      if IsConsole then
        WriteLn('''('' or ''}'' expected.');
      Exit
    end;

    case Token of
      BrckOp:
        if not GetTuple(Rails) then
          Exit;
      CrlBrckCl:
        Break;
    else
      begin
        if IsConsole then
          WriteLn(Format('Unexpected ''%s''', [Value]));
        Exit
      end
    end
  end;

  Result := True
end;

function GetDepths(var Depths: TDepths): Boolean;
begin
  Result := False;

  if (not GetNextSymbol(Token, Value)) or (Token <> BrckOp) then
  begin
    if IsConsole then
      WriteLn('''('' expected.');
    Exit
  end;

  while True do
  begin
    if not GetNextSymbol(Token, Value) then
    begin
      if IsConsole then
        WriteLn('Numeral or '')'' expected.');
      Exit
    end;
    case Token of
      Numeral:
        begin
          SetLength(Depths, Length(Depths) + 1);
          Depths[High(Depths)] := StrToInt(Value);
          if (not GetNextSymbol(Token, Value)) or ((Token <> Comma) and
            (Token <> BrckCl)) then
          begin
            if IsConsole then
              WriteLn(''','' or '')'' expected.');
            Exit
          end;
          if Token = BrckCl then
            Break
        end;
      BrckCl:
        Break
    else
      if IsConsole then
        WriteLn('Numeral or '')'' expected.');
      Exit
    end
  end;

  Result := True
end;

function GetLiteralStr(var Return: string): Boolean;
begin
  Result := False;

  if (not GetNextSymbol(Token, Value)) or (Token <> Colon) then
  begin
    if IsConsole then
      WriteLn(''':'' expected.');
    Exit
  end;

  if (not GetNextSymbol(Token, Value)) or (Token <> LiteralStr) then
  begin
    if IsConsole then
      WriteLn('Literal string expected.');
    Exit
  end;

  Return := Value;
  Result := True
end;

function GetNumeral(var Return: Longint): Boolean;
begin
  Result := False;

  if (not GetNextSymbol(Token, Value)) or (Token <> Colon) then
  begin
    if IsConsole then
      WriteLn(''':'' expected.');
    Exit
  end;

  if (not GetNextSymbol(Token, Value)) or (Token <> Numeral) then
  begin
    if IsConsole then
      WriteLn('Literal string expected.');
    Exit
  end;

  Result := TryStrToInt(Value, Return)
end;

function GetHotspot(var Hotspots: THotspots): Boolean;
var
  Hotspot: THotspot;
begin
  Result := False;

  // initialize data
  with Hotspot do
  begin
    Name := '';
    Verb := '';
    Prep := '';
    Feet_x := -1;
    Feet_y := -1;
    Syntax := -1;
    Facing := -1;
    Cursor := -1
  end;

  if (not GetNextSymbol(Token, Value)) or (Token <> Numeral) then
  begin
    if IsConsole then
      WriteLn('Numeral expected.');
    Exit;
  end;
  Hotspot.x1 := StrToInt(Value);
  if (not GetNextSymbol(Token, Value)) or (Token <> Comma) then
  begin
    if IsConsole then
      WriteLn(''','' expected.');
    Exit;
  end;

  if (not GetNextSymbol(Token, Value)) or (Token <> Numeral) then
  begin
    if IsConsole then
      WriteLn('Numeral expected.');
    Exit;
  end;
  Hotspot.y1 := StrToInt(Value);
  if (not GetNextSymbol(Token, Value)) or (Token <> Comma) then
  begin
    if IsConsole then
      WriteLn(''','' expected.');
    Exit;
  end;

  if (not GetNextSymbol(Token, Value)) or (Token <> Numeral) then
  begin
    if IsConsole then
      WriteLn('Numeral expected.');
    Exit;
  end;
  Hotspot.x2 := StrToInt(Value);
  if (not GetNextSymbol(Token, Value)) or (Token <> Comma) then
  begin
    if IsConsole then
      WriteLn(''','' expected.');
    Exit;
  end;

  if (not GetNextSymbol(Token, Value)) or (Token <> Numeral) then
  begin
    if IsConsole then
      WriteLn('Numeral expected.');
    Exit;
  end;
  Hotspot.y2 := StrToInt(Value);
  if (not GetNextSymbol(Token, Value)) or ((Token <> Comma)
    and (Token <> CrlBrckCl)) then
  begin
    if IsConsole then
      WriteLn(''','' or ''}'' expected.');
    Exit;
  end;

  if Token = CrlBrckCl then
  begin
    SetLength(Hotspots, Length(Hotspots) + 1);
    Hotspots[High(Hotspots)] := Hotspot;
    Result := True;
    Exit
  end;

  while GetNextSymbol(Token, Value) do
  begin
    case Token of
      Identifier:
        begin
          if Value = 'name' then
          begin
            if Hotspot.Name <> '' then
            begin
              if IsConsole then
                WriteLn('''name'' redefined.');
              Exit
            end;
            if not GetLiteralStr(Hotspot.Name) then
              Exit
          end
          else if Value = 'sprite' then
          begin
            if Hotspot.Sprite <> '' then
            begin
              if IsConsole then
                WriteLn('''sprite'' redefined.');
              Exit
            end;
            if not GetLiteralStr(Hotspot.Sprite) then
              Exit
          end
          else if Value = 'verb' then
          begin
            if Hotspot.Verb <> '' then
            begin
              if IsConsole then
                WriteLn('''verb'' redefined.');
              Exit
            end;
            if not GetLiteralStr(Hotspot.Verb) then
              Exit
          end
          else if Value = 'prep' then
          begin
            if Hotspot.Prep <> '' then
            begin
              if IsConsole then
                WriteLn('''prep'' redefined.');
              Exit
            end;
            if not GetLiteralStr(Hotspot.Prep) then
              Exit
          end
          else if Value = 'feet_x' then
          begin
            if Hotspot.Feet_x <> -1 then
            begin
              if IsConsole then
                WriteLn('''feet_x'' redefined.');
              Exit
            end;
            if not GetNumeral(Hotspot.Feet_x) then
              Exit
          end
          else if Value = 'feet_y' then
          begin
            if Hotspot.Feet_y <> -1 then
            begin
              if IsConsole then
                WriteLn('''feet_y'' redefined.');
              Exit
            end;
            if not GetNumeral(Hotspot.Feet_y) then
              Exit
          end
          else if Value = 'syntax' then
          begin
            if Hotspot.Syntax <> -1 then
            begin
              if IsConsole then
                WriteLn('''syntax'' redefined.');
              Exit
            end;
            if not GetNumeral(Hotspot.Syntax) then
              Exit
          end
          else if Value = 'facing' then
          begin
            if Hotspot.Facing <> -1 then
            begin
              if IsConsole then
                WriteLn('''facing'' redefined.');
              Exit
            end;
            if not GetNumeral(Hotspot.Facing) then
              Exit
          end
          else if Value = 'cursor' then
          begin
            if Hotspot.Cursor <> -1 then
            begin
              if IsConsole then
                WriteLn('''cursor'' redefined.');
              Exit
            end;
            if not GetNumeral(Hotspot.Cursor) then
              Exit
          end
          else
          begin
            if IsConsole then
              WriteLn(Format('Unexpected identifier ''%s''.', [Value]));
            Exit
          end;

          if (not GetNextSymbol(Token, Value)) or ((Token <> Comma) and
            (Token <> CrlBrckCl)) then
          begin
            if IsConsole then
              WriteLn(''','' or ''}'' expected.');
            Exit
          end;

          if Token = CrlBrckCl then
            Break
        end
    end
  end;

  SetLength(Hotspots, Length(Hotspots) + 1);
  Hotspots[High(Hotspots)] := Hotspot;

  Result := True
end;

function GetHotspots(var Hotspots: THotspots): Boolean;
begin
  Result := False;

  if not (GetNextSymbol(Token, Value)) or (Token <> CrlBrckOp) then
  begin
    if IsConsole then
      WriteLn('''{'' expected.');
    Exit
  end;

  while True do
  begin
    if not GetNextSymbol(Token, Value) then
    begin
      if IsConsole then
        WriteLn('Numeral or ''}'' expected.');
      Exit
    end;
    case Token of
      CrlBrckOp:
        if not GetHotspot(Hotspots) then
          Exit;
      CrlBrckCl:
        Break
    else
      if IsConsole then
        WriteLn('Numeral or ''}'' expected.');
      Exit
    end
  end;

  Result := True
end;

function Parse(var DEF: TDEF): Longint;
begin
  Result := DEFParser_ParserError;

  if not GetNextSymbol(Token, Value)then
  begin
    if (Token = Eoi) and IsConsole then
      WriteLn('Empty input.');
    Exit
  end;

  if Value <> 'scenefile' then
  begin
    if IsConsole then
      WriteLn(Format('''scenefile'' expected, ''%s'' found.', [Value]));
    Exit
  end;

  if not GetNextSymbol(Token, Value)then
  begin
    if (Token = Eoi) and IsConsole then
      WriteLn('Premature end of input, ''v5_00'' expected.');
    Exit
  end;

  if Value <> 'v5_00' then
  begin
    if IsConsole then
      WriteLn(Format('''v5_00'' expected, ''%s'' found.', [Value]));
    Exit
  end;

  // initialize information
  with DEF do
  begin
    Scene := '';
    Artbase := '';
    Frontscale := -1;
    Backscale := -1;
    Fronty := -1;
    Backy := -1;
    SetLength(Rails, 0);
    SetLength(Depths, 0);
    SetLength(Hotspots, 0)
  end;

  while GetNextSymbol(Token, Value) do
  begin
    case Token of
      Identifier:
        if Value = 'scene' then
        begin
          if not GetScene(DEF.Scene) then
            Exit
        end
        else if Value = 'artbase' then
        begin
          if not GetArtbase(DEF.Artbase) then
            Exit
        end
        else if Value = 'frontscale' then
        begin
          if not GetFrontscale(DEF.Frontscale) then
            Exit
        end
        else if Value = 'backscale' then
        begin
          if not GetBackscale(DEF.Backscale) then
            Exit
        end
        else if Value = 'fronty' then
        begin
          if not GetFronty(DEF.Fronty) then
            Exit
        end
        else if Value = 'backy' then
        begin
          if not GetBacky(DEF.Backy) then
            Exit
        end
        else if Value = 'rails' then
        begin
          if not GetRails(DEF.Rails) then
            Exit
        end
        else if Value = 'depths' then
        begin
          if not GetDepths(DEF.Depths) then
            Exit
        end
        else if Value = 'hotspots' then
        begin
          if not GetHotspots(DEF.Hotspots) then
            Exit
        end
        else
        begin
          if IsConsole then
            WriteLn(Format('Unknown identifier ''%s''.', [Value]));
          Exit
        end
    end
  end;

  Result := DEFParser_Success
end;

procedure CloseDEF;
begin
  CloseFile(DEFFile)
end;

function ParseDEF(DEFName: string; var DEF: TDEF): Longint;
begin
  Result := OpenDEF(DEFName);
  if Result <> DEFParser_Success then
    Exit;

  Result := Parse(DEF);

  CloseDEF
end;

end.
