{ File   : TTLoader.pas
  Author : Deniz Oezmen
  Created: 2005-03-07, based on TT2BMP.dpr created 2005-02-09
  Changed: 2006-08-13

  TT-to-TBitmap converter.
}
unit TTLoader;

interface

uses
  Graphics, Dialogs;

const
  // return codes
  TTLoader_Success     = Longint($00000000);
  TTLoader_FileNotOpen = Longint($80020001);
  TTLoader_InvalidFile = Longint($80020002);
  TTLoader_LoadError   = Longint($80020003);

function LoadTT(TTName: string; var Bitmap: TBitmap): Longint;

implementation

uses
  SysUtils;

type
  TColor   = packed record
               B, G, R, A: Byte
             end;

  TPalette = array[0..255] of TColor;

  THeader  = packed record
               Magic,      // $54542020 = '  TT' [#032#032#084#084]
               DataSize,
               XRes,
               YRes,
               XBlocks,
               YBlocks,
               XBlockSize,
               YBlockSize: Longint;
               Palette   : TPalette
             end;

var
  TTFile : file;
  Header : THeader;

function OpenTTFile(var TTFile: file; TTName: string): Longint;
begin
  FileMode := fmShareDenyWrite;
  AssignFile(TTFile, TTName);

  {$I-}
  Reset(TTFile, 1);
  {$I+}
  if IOResult = 0 then
    OpenTTFile := TTLoader_Success
  else
    OpenTTFile := TTLoader_FileNotOpen
end;

function ReadHeader(var Header: THeader): Longint;
var
  BytesRead: Longint;
begin
  BlockRead(TTFile, Header, SizeOf(Header), BytesRead);
  if (Header.Magic = $54542020) and (BytesRead = SizeOf(Header)) then
    ReadHeader := TTLoader_Success
  else
    ReadHeader := TTLoader_InvalidFile
end;

procedure ConvertPalette(var Palette: TPalette);
var
  i: Byte;
begin
  for i := 0 to 255 do
    with Palette[i] do
    begin
      R := R shl 2;
      G := G shl 2;
      B := B shl 2
    end
end;

procedure CopyBlock(Block, TT: Pointer; x, y: Longint; Header: THeader);
var
  SourceP,
  DestP,
  XSize,
  YSize,
  i      : Longint;
begin
  if x + Header.XBlockSize <= Header.XRes then
    XSize := Header.XBlockSize
  else
    XSize := Header.XRes mod Header.XBlockSize;
  if y + Header.YBlockSize <= Header.YRes then
    YSize := Header.YBlockSize
  else
    YSize := Header.YRes mod Header.YBlockSize;
  for i := 0 to YSize - 1 do
  begin
    SourceP := i * XSize;
    DestP := (y + i) * Header.XRes + x;
    Move(Pointer(Longint(Block) + SourceP)^,
      Pointer(Longint(TT) + DestP)^,
      XSize)
  end
end;

function ConvertToBMP(Header: THeader; var Bitmap: TBitmap): Longint;
var
  x, y : Longint;
  TT,
  Block: Pointer;
  Color: TColor;
begin
  Result := TTLoader_LoadError;

  GetMem(TT, Header.XRes * Header.YRes);
  GetMem(Block, Header.XBlockSize * Header.YBlockSize);

  for y := 0 to Header.YBlocks - 1 do
    for x := 0 to Header.XBlocks - 1 do
    begin
      {$I-}
      BlockRead(TTFile, Block^, Header.XBlockSize * Header.YBlockSize);
      {$I+}
      if IOResult <> 0 then
        Exit;
      CopyBlock(Block, TT, x * Header.XBlockSize, y * Header.YBlockSize, Header)
    end;

  FreeMem(Block, Header.XBlockSize * Header.YBlockSize);

  Bitmap.Width := Header.XRes;
  Bitmap.Height := Header.YRes;

  for y := 0 to Header.YRes - 1 do
    for x := 0 to Header.XRes - 1 do
    begin
      Color := Header.Palette[Byte(Pointer(Longint(TT) +
        y * Header.XRes + x)^)];
      Bitmap.Canvas.Pixels[x, y] := Color.B shl 16 + Color.G shl 8 + Color.R
    end;

  FreeMem(TT, Header.XRes * Header.YRes);

  Result := TTLoader_Success
end;

procedure CloseTTFile;
begin
  CloseFile(TTFile)
end;

function LoadTT(TTName: string; var Bitmap: TBitmap): Longint;
begin
  Result := OpenTTFile(TTFile, TTName);
  if Result <> TTLoader_Success then
    Exit;

  Result := ReadHeader(Header);
  if Result <> TTLoader_Success then
  begin
    CloseTTFile;
    Exit
  end;

  ConvertPalette(Header.Palette);

  Result := ConvertToBMP(Header, Bitmap);

  CloseTTFile
end;

end.


